import ModuleSystem from './ModuleSystem';
import ServiceSystem from './ServiceSystem';
import LinkManager from './LinkManager';
import RoutingSystem from './RoutingSystem';

import Systems from './Systems';
import Modules from '../modules';


class CoreSystem {
	constructor(id = 'core') {
		this.id = id;
		this.ModuleSystem = new ModuleSystem(id);
		this.ServiceSystem = new ServiceSystem(id);
		this.LinkManager = new LinkManager(id);
		this.RoutingSystem = new RoutingSystem(id);

		this.root = "root";
		this.RoutingSystem.onRouteChange("CoreSystem", () => {
			console.log("CHANGE" , this.RoutingSystem.getRoute());
			this.refresh(true, false);
			if(this._onUpdate) this._onUpdate();
		});

		this.inputs = {};
		this.outputs = {};
		this.inrefs = {};
		this.outrefs = {};
	}

	import(data) {
		let {
			services,
			modules,
			links,
			routes,
			outputs,
			inputs,
			root
		} = data;

		this.root = root || this.root;
		this.ServiceSystem.clean();
		this.ModuleSystem.clean();
		this.LinkManager.clean();
		this.RoutingSystem.clean();


		for(var i in inputs) {
			this.addInput(inputs[i]);
		}

		for(var i in outputs) {
			this.addOutput(outputs[i]);
		}

		for(var i in services) {
			this.ServiceSystem.loadService(services[i]);
		}
		for(var i in modules) {
			this.ModuleSystem.loadModule(modules[i]);
		}
		for(var i in links) {
			this.LinkManager.addLink(links[i]);
		}
		this.RoutingSystem.import(routes);
		
		this.refresh(true);
	}

	export() {
		let services = this.ServiceSystem.export(),
			modules = this.ModuleSystem.export(),
			links = this.LinkManager.export(),
			routes = this.RoutingSystem.export(),
			outputs = this.outputs,
			inputs = this.inputs;

		return {
			services,
			modules,
			links,
			routes,
			outputs,
			inputs,
			root: this.root
		};
	}

	addInput(input) {
		input.pid = 'core';
		this.inputs[input.name] = input;
		this.inrefs[input.name] = this.LinkManager.addInputRef(input);
		this.refresh();
	}
	addOutput(output) {
		output.pid = 'core';
		this.outputs[output.name] = output;
		this.outrefs[output.name] = this.LinkManager.addOutputRef(output);
		this.refresh();
	}
	updateInput(input) {
		input.pid = 'core';
		this.inputs[input.name] = input;
		this.refresh();
	}
	updateOutput(output) {
		output.pid = 'core';
		this.outputs[output.name] = output;
		this.refresh();
	}
	removeInput(name) { delete this.inputs[name];this.refresh(true); }
	removeOutput(name) { delete this.outputs[name];this.refresh(true); }
	getInput(name) {  return  this.inrefs[name]; }
	getOutput(name) { return this.outrefs[name]; }


	getLoadedModules() {
		return {...this.ModuleSystem.availableModules};
	}
	loadModule(mod) {
		this.ModuleSystem.loadModule(mod);
		this.refresh();
	}
	loadService(mod) {
		this.ServiceSystem.loadService(mod);
		this.refresh();
	}
	refresh(hard = false, suppresedCallback = false) {
		this.ModuleSystem.refresh(hard, true);
		this.ServiceSystem.refresh(hard, true);
		this.LinkManager.clearLinks({...this.ModuleSystem.modules, ...this.ServiceSystem.modules});
		if(this._onUpdate && !suppresedCallback) this._onUpdate();
	}

	onUpdate(fn) {
		this.ServiceSystem.onUpdate(fn);
		this.ModuleSystem.onUpdate(fn);
		this._onUpdate = fn;
	}

	clean() {
		this.ModuleSystem.clean();
		this.ServiceSystem.clean();
		this.LinkManager.clean();
		this.RoutingSystem.clean();
	}

	getRandom() {
		let r = parseInt(Math.random() * 1234567890) % 7654321;
		// check modules
		// check services 
		return r;
	}

	go(url, state, replace = false) {
		if(!this.RoutingSystem.goName(url)) {
			this.RoutingSystem.go(url, state, replace);
			if(replace)this.refresh();
		}
		this._routerState = state;
		//this.refresh(true);
	}

	render() {
		this.RoutingSystem.resolve();
		this.root = this.RoutingSystem.getCurrentView();
		return [this.ServiceSystem.render(), this.ModuleSystem.render(this.root)];
	}

	renderModules() {
		this.RoutingSystem.resolve();
		this.root = this.RoutingSystem.getCurrentView(); // maybe i should take Routing System out of here beacause of subSystems
		return this.ModuleSystem.render(this.root);
	}

	renderServices() {
		return this.ServiceSystem.render();
	}
}

export default CoreSystem;