import React from 'react';
import Logger from './Logger';
import { Resolver } from '../modules/ModuleUtilities';
import Systems from './Systems';

class ModuleSystem {
	constructor(systemID) {
		this.systemID = systemID;
		this.clean();
		this.availableModules = Systems.getModules();
	}

	clean() {
		this.modules = {};
		this.noresolve = {};
		this.resolved = {};
		this.metamodules = {};
	}

	ghostModule(id, ctor, inputs) {
		if (this.resolved[id]) {
			return {
				id: id,
				remove: () => this.remove(id),
				render: () => this.render(id)				
			}
		}

		let ins = {};
		for(var i in inputs){ 
			ins[i] = { value: inputs[i] }
		}
		let mod = {
			ctor: ctor,
			id: id,
			inputs: ins,
			readOnly: true
		};
		this.loadModule(mod);
		this.refresh(true);
		return {
			id: mod.id,
			remove: () => this.remove(mod.id),
			render: () => this.render(mod.id)
		}
	}

	loadModule(mod, rootMod = false) {
		mod.inputs = mod.inputs || {};
		mod.outputs = mod.outputs || {};
		mod.children = mod.children || [];

		let {
			ctor,
			id,
			inputs,
			outputs,
			children,
			readOnly
		} = mod;

		if(this.availableModules[ctor] === undefined) {
			Logger.warning("Module " + ctor + " is not bundled! You either forgot to add it to ModuleSystem or it doesn't exist!");
			return;
		}

		let inrefs = [];
		for(var i in this.availableModules[ctor].Inputs) {
			let v = this.availableModules[ctor].Inputs[i].defaultValue;
			if(inputs[i]) v = inputs[i].value || v;
			inputs[i] = {
				name: i,
				value: v,
				pid: id,
				arrType: this.availableModules[ctor].Inputs[i].arrType,
				type: this.availableModules[ctor].Inputs[i].type
			};
			inrefs[i] = this.LinkManager.addInputRef(inputs[i]);
		}
		let outrefs = [];
		for(var i in this.availableModules[ctor].Outputs) {
			outputs[i] = {
				name: i,
				pid: id,
				arrType: this.availableModules[ctor].Outputs[i].arrType,
				type: this.availableModules[ctor].Outputs[i].type
			};
			outrefs[i] = this.LinkManager.addOutputRef(outputs[i]);
		}

		this.__addToScene({...mod, children: [...children], inrefs, outrefs, key:''+id+id});
	}

	connect(input, output) {
		this.LinkManager.createLink(input, output);
	}

	rename(oldid, newid) {
		for (var i in this.metamodules) {
			// only first child
			let inx = this.metamodules[i].children.indexOf(oldid);
			if(inx > -1) {
				this.metamodules[i].children[inx] = newid;
			}
		}
		for(var i in this.metamodules[oldid].inrefs){
			this.metamodules[oldid].inputs[i].pid = newid;
			this.metamodules[oldid].inrefs[i].pid = newid;
		}
		for(var i in this.metamodules[oldid].outrefs){
			this.metamodules[oldid].outputs[i].pid = newid;
			this.metamodules[oldid].outrefs[i].pid = newid;
		}

		this.LinkManager.rename(oldid, newid);
		this.metamodules[newid] = this.metamodules[oldid];
		this.metamodules[newid].id = newid;
		delete this.metamodules[oldid];
		this._shuffleKeys(this.resolved[oldid]);
		delete this.resolved[oldid];
		this.noresolve = {...this.metamodules};
		this.modules = {};
	}

	reorder(modid, times) {
		let parent = this.resolved[modid].parent;
		let childIndx = parent.children.indexOf(modid);
		let swapIndx = childIndx + times;
		if ( childIndx > -1  && swapIndx > -1 && swapIndx < parent.children.length ) {
			let temp = parent.children[swapIndx];
			parent.children[swapIndx] = modid;
			parent.children[childIndx] = temp;
			this._shuffleKeys(this.resolved[modid]);
			this._shuffleKeys(this.resolved[temp]);
			this.noresolve = {...this.metamodules};
			this.modules = {};
		}
	}

	change(id, name, value) {
		let v = value;
		this.resolved[id].inputs[name].value = v;
		this.resolved[id].inrefs[name].value = v;
		// re-render
		//this._shuffleKey(this.metamodules[id]);
		//this._shuffleKeys(this.resolved[id]);
		//this.noresolve = {...this.metamodules};
		//this.modules = {};
	}

	addChild(id, childid) {
		this.resolved[id].children.push(childid);
		//this._shuffleKeys(this.resolved[id]);
		this.modules = {};
		this.noresolve = {...this.metamodules};
	}

	remove(id) {
		this.modules = {};
		let parent = this.resolved[id].parent;
		//this._shuffleKeys(this.resolved[id]);
		delete this.metamodules[id];
		if(parent){
			let indx = parent.children.indexOf(id);
			if(indx > -1)parent.children.splice(indx, 1);
		}
		this.noresolve = {...this.metamodules};
		this.resolved = {};
		this.LinkManager.breakLinks(id);
	}


	_shuffleAllKeys(temp) {
		this.modules = {};
		this.noresolve = {...this.metamodules};
		// for(var i in this.metamodules) {
		// 	this.metamodules[i].key = this.metamodules[i].id + Math.random() %231231;
		// }
	}

	_shuffleKeys(temp) {
		let i = 0;
		do{
			temp.key = temp.id + Math.random() %231231;
			temp = temp.parent;
			i++;
		} while(temp && i < 2); // i < 2 for only change parent
	}
	_shuffleKey(temp) {
		temp.key = temp.id + Math.random() %231231;
	}
	__Factory(item) {
		let children = [];
		for (var i=0;i<item.children.length;i++) {
			children.push(this.modules[item.children[i]]);
			this.resolved[item.children[i]].parent = item;
		}
		let ctor = this.availableModules[item.ctor];
		let props = item;
		props.System = this; // I have systemID but this is parametric for Module/Service
		props.systemID = this.systemID;
		props.decorate = function(systemid = "core", item) {
			Systems.decorate(systemid, item);
		}
		props.__adminmode = 'normal';
		return React.createElement(ctor, props, children);
	}

	__childrenReady(children) {
		for (var i=0;i<children.length;i++) {
			let id = children[i];
			if(!this.modules[id]) {
				return false;
			}
		}
		return true;
	}

	__resolveItem(item) {
		if (this.LinkManager.isReady(item) && this.__childrenReady(item.children)) {
			this.resolved[item.id] = item;
			let mod = this.__Factory(item);
			item.reactElement = mod;
			delete this.modules[item.id];
			this.modules[item.id] = mod;
			delete this.noresolve[item.id];
			this.dirty = true;
		}
	}

	__resolve() {
		for (var i in this.noresolve) {
			this.__resolveItem(this.noresolve[i]);
		}
	}

	__addToScene(item) {
		this.noresolve[item.id] = item;
		this.metamodules[item.id] = item;
	}

	__refresh(hard = false, suppresedCallbacks = false) {
		if(hard)this._shuffleAllKeys();

		do {
			this.dirty = false;
			this.__resolve();
		} while(this.dirty);
		if(this._onUpdate && !suppresedCallbacks)this._onUpdate();
	}
	refresh(h, s) { this.__refresh(h, s); }

	onUpdate(fn) {
		this._onUpdate = fn;
	}

	render(id) {
		let mod = this.modules[id || this.root];
		if(!mod) {
			mod = this.modules[this.root];
		}
		return mod || 'No Root';
	}

	getModules() {
		return Object.keys(this.resolved).map((key) => this.resolved[key]);
	}
	
	has(id) { return this.modules[id]; }

	export() {
		return Object.keys(this.resolved).map((key) => {
			let item = this.resolved[key];
			return {
				ctor: item.ctor,
				children: item.children,
				id: item.id,
				inputs: item.inputs,
				outputs: item.outputs,
				key: item.key,
				readOnly: item.readOnly
			};
		});
	}

}

export default ModuleSystem;